"""file: closest_pair.py --

Implements the closest pair algorithm, as question 6 of the homework.

"""

from mediane import median
import sort
import math

# here is a trick: in python any integer is strictly less that any
# string, so we can define an infinity as follow
INF = ''

def dist(p,q):
  """p,q are couple of floats

  computes the distance between points p and q

  """
  (x0,y0),(x1,y1) = p,q
  x,y = x1-x0,y1-y0
  return math.sqrt(x*x+y*y)

def naive(cloud):
  """cloud is a list of points

  computes the smallest distance between distinct points of 'cloud' in
  the most naive way: by testing all possible couples

  """
  n = len(cloud)
  m = INF
  for i in range(n):
    for j in range(i+1,n):
      d = dist(cloud[i],cloud[j])
      if d<m: m = d
  return m

def split(cloud):
  """cloud is a list of points

  splits the cloud according to the x-median 

  """
  z = median(cloud,lambda p,q: p[0]<=q[0])[0]
  left,right = [],[]
  for p in cloud:
    (x,_) = p
    if x <= z: left.append(p)
    else: right.append(p)
  return (left,right,z)

def splitzone(left,right,z,d):
  """left,right are lists of points; z is a float; d is a float

  left and right are supposed to be y-sorted.

  returns the minimal distance between distincts points of left+right
  that are also in the zone between x=z-d and x=z+d. It is done in
  linear time.

  """
  n1,n2 = len(left),len(right)
  zone = []
  # custom merge
  i,j = 0,0
  while i<n1 and j<n2:
    (x1,y1),(x2,y2) = left[i],right[j]
    if y1<=y2:
      i += 1
      if x1>z-d: zone.append((x1,y1))
    else:
      j += 1
      if x2<z+d: zone.append((x2,y2))
  while i<n1:
    (x1,y1) = left[i]
    i += 1
    if x1>z-d: zone.append((x1,y1))
  while j<n2:
    (x2,y2) = right[j]
    j += 1
    if x2<z+d: zone.append((x2,y2))
  # walk through zone
  m = INF
  i = 0
  while i<len(zone)-1:
    (_,y) = zone[i]
    j = i+1
    (_,yy) = zone[j]
    # the following loop makes a bounded (by 9) number of steps
    while j<len(zone) and yy<=y+d:
      dd = dist(zone[i],zone[j])
      if dd<m: m = dd
      j += 1
    i += 1
  return m

def closest_rec(cloud):
  """cloud is a list of points, supposed to be y-sorted
  
  computes the minimal distance between distinct points of cloud with
  a divide-and-conquer strategy: split cloud according to a x-median
  point, recursively tackle the task on the left and right subcloud,
  then compare the two results with the minimum distance in the 'zone'
  as defined in the homework

  """
  if len(cloud) < 2: return INF
  if len(cloud) == 2: return dist(cloud[0],cloud[1])
  else:
    (left,right,z) = split(cloud)
    d1,d2 = closest_rec(left),closest_rec(right)
    d = splitzone(left,right,z,min(d1,d2))
    return min(d1,d2,d)

def closest(cloud):
  """cloud is a list of more than two points

  returns the minimal distance between distinct points, by first
  y-sorting it and then giving it to closest_rec to process.

  """
  assert len(cloud)>1
  return closest_rec(sort.heapsort(cloud,lambda p,q: p[1]<=q[1]))
  
## testing area
import random
ntest,minsize,maxsize = 1000,20,50
swcorner,necorner = -15.,50.
nofail = True
for i in xrange(ntest):
  ncloud = random.randrange(minsize,maxsize)
  x = [(random.uniform(swcorner,necorner),random.uniform(swcorner,necorner))
   for i in range(ncloud)]
  y1,y2 = naive(x), closest(x)
  if not y1==y2:
    print "fail! on size:", ncloud, "values:", naive(x), closest(x)
if nofail: print "All", ntest," test(s) passed successfully."
