"""file: mediane.py

This file implements the course's method to find the median element of
a list in linear time.

"""

import sort

def naive_median(li,comp):
  """li is a list of integers

  Returns a median element in a naive fashion, by first sorting the
  list according to the comparison function 'comp'. Will be used on
  small lists.

  """
  return sort.heapsort(li,comp)[len(li)/2]

def median_approx(li,sbatch,comp):
  """li is a list of integers

  Returns an approximative median element in linear time. It divides
  the list into small batches of sizes sbatch on which the median is
  computed naively, then recursively computes the median of all those
  local median values.

  """
  n = len(li)
  assert n>0 #li must be non empty
  if n==1: return li[0]
  else:
    return median_approx([naive_median(li[i:i+sbatch],comp) for i in
                          range(0,n,sbatch)],sbatch,comp)

def select(li,i,comp):
  """li is a list of integers, i is an index

  Returns the i-th element of a list.

  """
  middle = median_approx(li,5,comp)
  inf,eq,sup = [],[],[]
  for x in li:
    if x == middle: eq.append(x)
    elif comp(x,middle): inf.append(x)
    else: sup.append(x)
  if i <= len(inf): return select(inf,i,comp)
  elif i <= len(inf)+len(eq): return middle
  else: return select(sup,i-len(inf)-len(eq),comp)

def median(li,comp):
  """li is a list of integers

  Returns a median element in linear time.

  """
  n = len(li)
  return select(li, n/2 + 1,comp)
